# -*- coding: utf-8 -*-
'''stuf has attributes'''

from stuf.core import defaultstuf, fixedstuf, frozenstuf, orderedstuf, stuf

__version__ = (0, 9, 0)

__all__ = 'defaultstuf fixedstuf frozenstuf orderedstuf stuf'.split()
